#'---
#'title: "Compilation Report | Corpus der Entscheidungen des Bundesverfassungsgerichts"
#'author: Seán Fobbe
#'papersize: a4
#'geometry: margin=3cm
#'fontsize: 11pt
#'output:
#'  pdf_document:
#'    toc: true
#'    toc_depth: 3
#'    number_sections: true
#'    pandoc_args: --listings
#'    includes:
#'      in_header: General_Source_TEX_Preamble_DE.tex
#'      before_body: [CE-BVerfG_Source_TEX_Definitions.tex,CE-BVerfG_Source_TEX_CompilationTitle.tex]
#'bibliography: packages.bib
#'nocite: '@*'
#' ---



#'\newpage
#+
#'# Einleitung
#'
#+
#'## Überblick
#' Dieses R-Skript lädt alle auf www.bundesverfassungsgericht.de verfügbaren Entscheidungen des Bundesverfassungsgerichts (BVerfG) herunter und verarbeitet sie in einen reichhaltigen menschen- und maschinenlesbaren Korpus. Es ist die Grundlage für den \textbf{\datatitle\ (\datashort )}.
#'
#' Alle mit diesem Skript erstellten Datensätze werden dauerhaft kostenlos und urheberrechtsfrei auf Zenodo, dem wissenschaftlichen Archiv des CERN, veröffentlicht. Alle Versionen sind mit einem persistenten Digital Object Identifier (DOI) versehen. Die neueste Version des Datensatzes ist immer über den Link der Concept DOI erreichbar: \dataconcepturldoi



#+
#'## Endprodukte

#' Primäre Endprodukte des Skripts sind folgende ZIP-Archive:
#' \begin{enumerate}
#' \item Der volle Datensatz im CSV-Format
#' \item Die reinen Metadaten im CSV-Format (wie unter 1, nur ohne Entscheidungstexte)
#' \item Der volle Datensatz im TXT-Format (reduzierter Umfang an Metadaten)
#' \item Der volle Datensatz im PDF-Format (reduzierter Umfang an Metadaten)
#' \item Alle Analyse-Ergebnisse (Tabellen als CSV, Grafiken als PDF und PNG)
#' \item Der Source Code und alle weiteren Quelldaten
#' \end{enumerate}
#'
#' Zusätzliche werden für alle ZIP-Archive kryptographische Signaturen (SHA2-256 und SHA3-512) berechnet und in einer CSV-Datei hinterlegt. Die Analyse-Ergebnisse werden zum Ende hin nicht gelöscht, damit sie für die Codebook-Erstellung verwendet werden können. Weiterhin kann optional ein PDF-Bericht erstellt werden (siehe unter "Kompilierung").


#+
#'## Systemanforderungen
#' Das Skript in seiner veröffentlichten Form kann nur unter Linux ausgeführt werden, da es Linux-spezifische Optimierungen (z.B. Fork Cluster) und Shell-Kommandos (z.B. OpenSSL) nutzt. Das Skript wurde unter Fedora Linux entwickelt und getestet. Die zur Kompilierung benutzte Version entnehmen Sie bitte dem **sessionInfo()**-Ausdruck am Ende dieses Berichts.
#'
#' In der Standard-Einstellung wird das Skript vollautomatisch die maximale Anzahl an Rechenkernen/Threads auf dem System zu nutzen. Wenn die Anzahl Threads (Variable "fullCores") auf 1 gesetzt wird, ist die Parallelisierung deaktiviert.
#'
#' Auf der Festplatte sollten 2 GB Speicherplatz vorhanden sein.
#' 
#' Um die PDF-Berichte kompilieren zu können benötigen Sie das R package **rmarkdown**, eine vollständige Installation von \LaTeX\ und alle in der Präambel-TEX-Datei angegebenen \LaTeX\ Packages.




#+
#'## Kompilierung

#' Mit der Funktion **render()** von **rmarkdown** können der **vollständige Datensatz** und das **Codebook** kompiliert und die Skripte mitsamt ihrer Rechenergebnisse in ein gut lesbares PDF-Format überführt werden.
#'
#' Alle Kommentare sind im roxygen2-Stil gehalten. Die beiden Skripte können daher auch **ohne render()** regulär als R-Skripte ausgeführt werden. Es wird in diesem Fall kein PDF-Bericht erstellt und Diagramme werden nicht abgespeichert.

#+
#'### Datensatz 

#' 
#' Um den vollständigen Datensatz zu kompilieren und einen PDF-Bericht zu erstellen, kopieren Sie bitte alle im Source-Archiv bereitgestellten Dateien in einen leeren Ordner und führen mit R diesen Befehl aus:

#+ eval = FALSE

rmarkdown::render(input = "CE-BVerfG_Source_CorpusCreation.R",
                  output_file = paste0("CE-BVerfG_",
                                       Sys.Date(),
                                       "_CompilationReport.pdf"),
                  envir = new.env())


#'### Codebook
#' Um das **Codebook** zu kompilieren und einen PDF-Bericht zu erstellen, kopieren Sie bitte alle im Source-Archiv bereitgestellten Dateien in einen leeren Ordner und führen im Anschluss an die Kompilierung des Datensatzes (!) untenstehenden Befehl mit R aus.
#'
#' Bei der Prüfung der GPG-Signatur wird ein Fehler auftreten und im Codebook dokumentiert, weil die Daten nicht mit meiner Original-Signatur versehen sind. Dieser Fehler hat jedoch keine Auswirkungen auf die Funktionalität und hindert die Kompilierung nicht.

#+ eval = FALSE

rmarkdown::render(input = "CE-BVerfG_Source_CodebookCreation.R",
                  output_file = paste0("CE-BVerfG_",
                                       Sys.Date(),
                                       "_Codebook.pdf"),
                  envir = new.env())





#'\newpage
#+
#'# Parameter

#+
#'## Name des Datensatzes
datasetname <- "CE-BVerfG"

#'## DOI des Datensatz-Konzeptes

doi.concept <- "10.5281/zenodo.3902658"

#'## DOI der konkreten Version

doi.version <- "10.5281/zenodo.4308215"


#'## Verzeichnis für Analyse-Ergebnisse
#' Hinweis: Muss mit einem Schrägstrich enden!
outputdir <- paste0(getwd(), "/ANALYSE/") 


#'## Optionen: Quanteda
tokens_locale <- "de_DE"


#'## Optionen: Knitr

#+
#'### Ausgabe-Format
dev <- c("pdf", "png")

#'### DPI für Raster-Grafiken
dpi <- 150

#'### Ausrichtung von Grafiken im Compilation Report
fig.align <- "center"


#'## Frequenztabellen: Ignorierte Variablen

#' Diese Variablen werden bei der Erstellung der Frequenztabellen nicht berücksichtigt.

varremove <- c("text",
               "eingangsnummer",
               "datum",
               "doc_id",
               "seite",
               "name",
               "ecli",
               "aktenzeichen")


#'# Vorbereitung

#'## Datumsstempel
#' Dieser Datumsstempel wird in alle Dateinamen eingefügt. Er wird am Anfang des Skripts gesetzt, für den den Fall, dass die Laufzeit die Datumsbarriere durchbricht.

datestamp <- Sys.Date()
print(datestamp)


#'## Datum und Uhrzeit (Beginn)

begin.script <- Sys.time()
print(begin.script)


#'## Ordner für Analyse-Ergebnisse erstellen
dir.create(outputdir)


#+
#'## Packages Laden

library(httr)         # HTTP-Werkzeuge
library(rvest)        # HTML/XML-Extraktion
library(knitr)        # Professionelles Reporting
library(kableExtra)   # Verbesserte Kable Tabellen
library(pdftools)     # Verarbeitung von PDF-Dateien
library(doParallel)   # Parallelisierung
library(ggplot2)      # Fortgeschrittene Datenvisualisierung
library(scales)       # Skalierung von Diagrammen
library(data.table)   # Fortgeschrittene Datenverarbeitung
library(readtext)     # TXT-Dateien einlesen
library(quanteda)     # Fortgeschrittene Computerlinguistik


#'## Zusätzliche Funktionen einlesen
#' **Hinweis:** Die hieraus verwendeten Funktionen werden jeweils vor der ersten Benutzung in vollem Umfang angezeigt um den Lesefluss zu verbessern.

source("General_Source_Functions.R")


#'## Quanteda-Optionen setzen
quanteda_options(tokens_locale = tokens_locale)


#'## Knitr Optionen setzen
knitr::opts_chunk$set(fig.path = outputdir,
                      dev = dev,
                      dpi = dpi,
                      fig.align = fig.align)



#'## Vollzitate statistischer Software
knitr::write_bib(c(.packages()), "packages.bib")


#'## Parallelisierung aktivieren
#' Parallelisierung wird zur Beschleunigung der Konvertierung von PDF zu TXT und der Datenanalyse mittels **quanteda** und **data.table** verwendet. Die Anzahl threads wird automatisch auf das verfügbare Maximum des Systems gesetzt, kann aber auch nach Belieben auf das eigene System angepasst werden. Die Parallelisierung kann deaktiviert werden, indem die Variable **fullCores** auf 1 gesetzt wird.
#'
#' Der Download der Dateien von \url{https://www.bundesverfassungsgericht.de} ist absichtlich nicht parallelisiert, damit das Skript nicht versehentlich als DoS-Tool verwendet wird.
#'
#' Die hier verwendete Funktion **makeForkCluster()** ist viel schneller als die Alternativen, funktioniert aber nur auf Unix-basierten Systemen (Linux, MacOS).

#+
#'### Logische Kerne

fullCores <- detectCores()
print(fullCores)

#'### Quanteda
quanteda_options(threads = fullCores) 

#+
#'### Data.table
setDTthreads(threads = fullCores)  

#'### DoParallel
cl <- makeForkCluster(fullCores)
registerDoParallel(cl)








#'# Download

#'## Zeitstempel: Linksammlung Beginn

begin.links <- Sys.time()
print(begin.links)


#'## Download vorbereiten

#+
#'### Funktion zeigen

print(f.linkextract)


#'### Maximale Seitenzahl auslesen

temp <- f.linkextract("https://www.bundesverfassungsgericht.de/SiteGlobals/Forms/Suche/Entscheidungensuche_Formular.html?language_=de")

temp1 <- grep("list.*253D([0-9]+).*",
              temp,
              value = TRUE)

temp2 <- gsub(".*253D([0-9]+).*",
              "\\1",
              temp1)

maxpage <- max(as.numeric(temp2))

print(maxpage)


#'### Linkliste erstellen
#' **Erstversuch**

pages <- seq_len(maxpage)
###pages <- sample(pages, 15) ### Begrenzung des Umfangs; nur für Debugging!
breaks <- sample(pages, 10)


range <- seq_along(pages)

links.list <- vector("list",
                     length(range))


print(pages)

for (i in range){

    URL  <- paste0("https://www.bundesverfassungsgericht.de/SiteGlobals/Forms/Suche/Entscheidungensuche_Formular.html?gtp=5403124_list%253D",
                   pages[i],
                   "&language_=de")
    
    volatile <- f.linkextract(URL)
    
    links.l1 <- grep ("SharedDocs/Entscheidungen",
                      volatile,
                      ignore.case = TRUE,
                      value = TRUE)

    links.list[[i]] <- links.l1
    Sys.sleep(runif(1, 0.8, 2))
    if (i %in% breaks) Sys.sleep(runif(1, 2, 7))
    
    }


Sys.sleep(runif(1, 20, 40))



#' **Zweitversuch**
#'
#'
#' Jede Seite der Datenbank sollte im Idealfall 10 Entscheidungen enthalten. Alle Seiten, die diese Bedingung nicht erfüllen werden noch einmal geprüft. Die letzte Seite enhält fast immer weniger als 10 Entscheidungen, wird sicherheitshalber aber trotzdem geprüft, statt sie auszusortieren. Dabei entstehende Duplikate werden später durch **unique()** entfernt.

retry.pages <- which(unlist(lapply(links.list,
                                   function(x)length(x)<10)))

retry.range <- seq_along(retry.pages)

retry.links.list <- vector("list",
                           length(retry.pages))


#' Anzeigen der Seiten die noch einmal geprüft werden.

print(retry.pages)



for (i in retry.range){

    URL  <- paste0("https://www.bundesverfassungsgericht.de/SiteGlobals/Forms/Suche/Entscheidungensuche_Formular.html?gtp=5403124_list%253D",
                   retry.pages[i],
                   "&language_=de")
    
    volatile <- f.linkextract(URL)
    
    links.l1 <- grep ("SharedDocs/Entscheidungen",
                      volatile,
                      ignore.case = TRUE,
                      value = TRUE)

    retry.links.list[[i]] <- links.l1
    
    Sys.sleep(runif(1, 0.8, 2))
    
    }


#' Anzeigen der Links die beim Wiederholungsversuch gesammelt wurden.

print(retry.links.list)


#'## Zeitstempel: Linksammlung Ende

end.links <- Sys.time()
print(end.links)


#'## Dauer: Linksammlung
end.links - begin.links


#'## HTML-Links in PDF-Links umwandeln

links.out <- c(unlist(links.list),
               unlist(retry.links.list))


links.pdf <- paste0("https://www.bundesverfassungsgericht.de/",
                    links.out)

links.pdf <- gsub("Entscheidungen",
                  "Downloads",
                  links.pdf)

links.pdf <- gsub("\\.html.*",
                  "\\.pdf\\?__blob=publicationFile\\&v\\=1",
                  links.pdf)


links.pdf <- unique(links.pdf)




#'## Dateinamen erstellen

#+
#'### Extrahieren relevanter Metadaten
#' Die Links zu jeder Entscheidung enthalten das Ordinalzahl-Element ihres jeweiligen ECLI-Codes. Struktur und Inhalt der ECLI für deutsche Gerichte sind auf dem Europäischen Justizportal näher erläutert. \footnote{\url{https://e-justice.europa.eu/content_european_case_law_identifier_ecli-175-de-de.do?member=1}}

filenames <- basename(links.pdf)

filenames <- gsub("[?].*",
                  "",
                  filenames)


#' **Normale Struktur**

filenames1 <- gsub("[a-z]([a-z])([0-9]{4})([0-9]{2})([0-9]{2})_([0-9])([a-z]*)([0-9]{4})([0-9]{2}).*",
                   "BVerfG_\\2-\\3-\\4_\\1_\\5_\\6_\\7_\\8_NA",
                   filenames)


#' **Struktur von Verzögerungsrügen**

filenames1 <- gsub("_vz", "NA_Vz", filenames1)

filenames1 <- gsub("[a-z]([a-z])([0-9]{4})([0-9]{2})([0-9]{2})(NA)_(Vz)([0-9]{4})([0-9]{2}).*",
                   "BVerfG_\\2-\\3-\\4_\\1_\\5_\\6_\\7_\\8_NA",
                   filenames1)


#' **Struktur von Entscheidungen mit Kollisions-Variable**

filenames1 <- gsub("[a-z]([a-z])([0-9]{4})([0-9]{2})([0-9]{2})([a-z0-9])_([0-9])([a-z]*)([0-9]{4})([0-9]{2}).*",
                   "BVerfG_\\2-\\3-\\4_\\1_\\6_\\7_\\8_\\9_\\5",
                   filenames1)




#'### Formatierung von Registerzeichen anpassen

filenames1 <- gsub("_bv([a-z])_",
                   "_Bv\\U\\1_",
                   perl = TRUE,
                   filenames1)


filenames1 <- gsub("pbvu",
                   "PBvU",
                   filenames1)


#'### Formatierung von Spruchkörper-Typ anpassen

filenames1 <- gsub("_([kpsb])_",
                   "_\\U\\1_",
                   perl = TRUE,
                   filenames1)




#'### Strenge REGEX-Validierung der Dateinamen
#' Das Ergebnis sollte ein leerer Vektor sein!

grep("BVerfG_[0-9]{4}-[0-9]{2}-[0-9]{2}_[A-Z]_[0-9NA]*_[A-Za-z]*_[0-9]{4}_[0-9]{2}_[0-9a-zNA]*$",
     filenames1,
     invert = TRUE,
     value = TRUE)




#'### Zusätzliche Variablen einfügen

extravariablen <- fread("C-BVerfGE_Source_Variablen_NameBandSeite.csv")

extravariablen$newname <- paste(extravariablen$oldname,
                                extravariablen$name,
                                extravariablen$band,
                                extravariablen$seite,
                                sep = "_")

extravariablen$newname <- paste0(extravariablen$newname,
                                 ".pdf")


filenames2 <- filenames1


targetindices <- match(extravariablen$oldname,
                       filenames2)

newname <- extravariablen$newname

dt <- data.table(targetindices, newname)[complete.cases(targetindices)]



if(dt[,.N] > 0){
    
filenames2 <- replace(filenames2,
                      dt$targetindices,
                      dt$newname)
}



#'### NAs einfügen für nicht benannte Entscheidungen

values <- grep(".pdf",
               filenames2,
               invert = TRUE,
               value = TRUE)

indices <- grep(".pdf",
                filenames2,
                invert = TRUE)

filenames2[indices] <- paste0(values,
                              "_NA_NA_NA.pdf")




#'### Strenge REGEX-Validierung der Dateinamen
#' Das Ergebnis sollte ein leerer Vektor sein!

grep("^BVerfG_[0-9]{4}-[0-9]{2}-[0-9]{2}_[SPKB]_[0-9NA]*_[A-Za-z]*_[0-9]{4}_[0-9]{2}_[0-9a-zNA]*_[0-9ÄÜÖäüöA-Za-z\\-]*_[NA0-9]*_[NA0-9]*\\.pdf$",
     filenames2,
     value=TRUE,
     invert=TRUE)





#'## Data Table für Download erstellen

dt <- data.table(links.pdf,
                 filenames2)


#'## Zeitstempel: Download Beginn

begin.download <- Sys.time()
print(begin.download)


#'## Download durchführen
#' **Hinweis:** Es ist nötig jeden Link auf das Vorhandensein einer PDF-Datei zu prüfen, weil für manche Entscheidungen zwar HTML-Seiten vorhanden sind, aber keine korrespondierende PDF-Datei.


for (i in seq_len(dt[,.N])){
    response <- GET(dt$links.pdf[i])
    Sys.sleep(runif(1, 0.25, 0.75))
    if (response$headers$"content-type" == "application/pdf;charset=UTF-8" & response$status_code == 200){
        tryCatch({download.file(url = dt$links.pdf[i],
                                destfile = dt$filenames2[i])
        },
        error=function(cond) {
            return(NA)}
        )     
    }else{
        print(paste0(dt$filenames2[i],
                     " : kein PDF vorhanden"))  
    }
    Sys.sleep(runif(1, 0.3, 1))
}   


#'## Zeitstempel: Download Ende

end.download <- Sys.time()
print(end.download)


#'## Dauer: Download 
end.download - begin.download


#'## Download: Ergebnis

#+
#'### Anzahl herunterzuladender Dateien
dt[,.N]

#'### Anzahl heruntergeladener Dateien
files.pdf <- list.files(pattern = "\\.pdf")
length(files.pdf)

#'### Fehlbetrag
N.missing <- dt[,.N] - length(files.pdf)
print(N.missing)

#'### Fehlende Dateien
missing <- setdiff(dt$filenames2,
                   files.pdf)

print(missing)





#'## Wiederholungsversuch
#' Download für fehlende Dokumente wiederholen.

if(N.missing > 0){

    dt.retry <- dt[filenames2 %in% missing]
    
    for (i in seq_len(dt.retry[,.N])){
        response <- GET(dt.retry$links.pdf[i])
        Sys.sleep(runif(1, 0.25, 0.75))
        if (response$headers$"content-type" == "application/pdf;charset=UTF-8" & response$status_code == 200){
            tryCatch({download.file(url = dt.retry$links.pdf[i],
                                    destfile = dt.retry$filenames2[i])
            },
            error=function(cond) {
                return(NA)}
            )     
        }else{
            print(paste0(dt.retry$filenames2[i],
                         " : kein PDF vorhanden"))  
        }
        Sys.sleep(runif(1, 0.5, 1.5))
    } 
}


#'## Download: Gesamtergebnis

#+
#'### Anzahl herunterzuladender Dateien
dt[,.N]

#'### Anzahl heruntergeladener Dateien
files.pdf <- list.files(pattern = "\\.pdf")
length(files.pdf)

#'### Fehlbetrag
N.missing <- dt[,.N] - length(files.pdf)
print(N.missing)

#'### Fehlende Dateien
missing <- setdiff(dt$filenames2, files.pdf)
print(missing)




#'## Abschließende Hinweise
#' **Hinweis:** Für die Entscheidung vom 1.10.2001 zur Rückgabe von EDV-Anlagen im Rahmen des NPD-Verfahrens war auch nach manueller Suche keine PDF-Datei auffindbar.






#+
#'# Text-Extraktion

#+
#'## Vektor der zu extrahierenden Dateien erstellen

files.pdf <- list.files(pattern = "\\.pdf$",
                        ignore.case = TRUE)


#'## Anzahl zu extrahierender Dateien
length(files.pdf)


#'## Seiten zählen: Funktion anzeigen
#+ results = "asis"
print(f.dopar.pagenums)


#'## Anzahl zu extrahierender Seiten
sum(f.dopar.pagenums(files.pdf))


#'## PDF extrahieren: Funktion anzeigen
#+ results = "asis"
print(f.dopar.pdfextract)


#'## Text Extrahieren
f.dopar.pdfextract(files.pdf)






#'# Korpus Erstellen

#+
#'## TXT-Dateien Einlesen

txt.bverfg <- readtext("./*.txt",
                    docvarsfrom = "filenames", 
                    docvarnames = c("gericht",
                                    "datum",
                                    "spruchkoerper_typ",
                                    "spruchkoerper_az",
                                    "registerzeichen",
                                    "eingangsnummer",
                                    "eingangsjahr_az",
                                    "kollision",
                                    "name",
                                    "band",
                                    "seite"),
                    dvsep = "_", 
                    encoding = "UTF-8")


#'## In Data Table umwandeln
setDT(txt.bverfg)


#'## Variable "datum" als Datentyp "IDate" kennzeichnen
txt.bverfg$datum <- as.IDate(txt.bverfg$datum)

#'## Datensatz nach Datum sortieren
#' Aufgrund der Position der Datums-Variable ist der Datensatz vermutlich schon von Linux nach Datum sortiert worden. Die Erstellung der Variablen für Präsidenten und Vize-Präsidenten trifft allerdings die starke Annahme, dass eine aufsteigende Sortierung nach Datum besteht. Wäre das nicht der Fall, würden dort Fehler auftreten. Diese Sortierung ist als fail-safe gedacht.

setorder(txt.bverfg, datum)


#'## Variable "entscheidungsjahr" hinzufügen
txt.bverfg$entscheidungsjahr <- year(txt.bverfg$datum)


#'## Variable "eingangsjahr_iso" hinzufügen
txt.bverfg$eingangsjahr_iso <- f.year.iso(txt.bverfg$eingangsjahr_az)






#'## Variable "praesi" hinzufügen
#' Diese Variable dokumentiert für jede Entscheidung welche/r PräsidentIn am Tag der Entscheidung im Amt war.

#+
#'### Lebensdaten einlesen

praesi <- fread("C-BVerfGE_Source_Variablen_Praesident.csv")

#'### Lebensdaten anzeigen

kable(praesi,
      format = "latex",
      align = "r",
      booktabs=TRUE,
      longtable=TRUE)



#'### Hypothethisches Amtsende für PräsidentIn
#' Weil der/die aktuelle PräsidentIn noch im Amt ist, ist der Wert für das Amtsende "NA". Dieser ist aber für die verwendete Logik nicht greifbar, weshalb an dieser Stelle ein hypothetisches Amtsende in einem Jahr ab dem Tag der Datensatzerstellung fingiert wird. Es wird nur an dieser Stelle verwendet und danach verworfen.

praesi[is.na(AmtszeitEnde)]$AmtszeitEnde <- Sys.Date() + 365


#'### Schleife vorbereiten

N <- praesi[,.N]

praesi.list <- vector("list", N)


#'### Vektor erstellen

for (i in seq_len(N)){
    praesi.N <- txt.bverfg[datum >= praesi$AmtszeitBeginn[i] & datum < praesi$AmtszeitEnde[i], .N]
    praesi.list[[i]] <- rep(praesi$Nachname[i],
                            praesi.N)
}



#'### Vektor einfügen

txt.bverfg$praesi <- unlist(praesi.list)



#'## Variable "v_praesi" hinzufügen
#' Diese Variable dokumentiert für jede Entscheidung welche/r Vize-PräsidentIn am Tag der Entscheidung im Amt war.


#+
#'### Lebensdaten einlesen

vpraesi <- fread("C-BVerfGE_Source_Variablen_VizePraesident.csv")

#'### Lebensdaten anzeigen

kable(vpraesi,
      format = "latex",
      align = "r",
      booktabs=TRUE,
      longtable=TRUE)


#'### Hypothethisches Amtsende für Vize-PräsidentIn
#' Weil der/die aktuelle Vize-PräsidentIn noch im Amt ist, ist der Wert für das Amtsende "NA". Dieser ist aber für die verwendete Logik nicht greifbar, weshalb an dieser Stelle ein hypothetisches Amtsende in einem Jahr ab dem Tag der Datensatzerstellung fingiert wird. Es wird nur an dieser Stelle verwendet und danach verworfen.

vpraesi[is.na(AmtszeitEnde)]$AmtszeitEnde <- Sys.Date() + 365


#'### Schleife vorbereiten

N <- vpraesi[,.N]

vpraesi.list <- vector("list", N)



#'### Vektor erstellen

for (i in seq_len(N)){
    vpraesi.N <- txt.bverfg[datum >= vpraesi$AmtszeitBeginn[i] & datum < vpraesi$AmtszeitEnde[i], .N]
    vpraesi.list[[i]] <- rep(vpraesi$Nachname[i],
                            vpraesi.N)
}



#'### Vektor einfügen

txt.bverfg$v_praesi <- unlist(vpraesi.list)







#'## Variable "aktenzeichen" hinzufügen

txt.bverfg$aktenzeichen <- paste0(txt.bverfg$spruchkoerper_az,
                                  " ",
                                  txt.bverfg$registerzeichen,
                                  " ",
                                  txt.bverfg$eingangsnummer,
                                  "/",
                                  txt.bverfg$eingangsjahr_az)

#' Bei Entscheidungen der Verzögerungskammer fehlt das Spruchkörper-Element des Aktenzeichens. Diese Zeile entfernt die "NA"-Angabe um ein korrektes Aktenzeichen herzustellen.

txt.bverfg$aktenzeichen <- gsub("NA ", "", txt.bverfg$aktenzeichen)




#'## Variable "ecli" hinzufügen
#' Struktur und Inhalt der ECLI für deutsche Gerichte sind auf dem Europäischen Justizportal näher erläutert. \footnote{\url{https://e-justice.europa.eu/content_european_case_law_identifier_ecli-175-de-de.do?member=1}}
#'
#' Sofern die Variablen korrekt extrahiert wurden lässt sich die ECLI vollständig rekonstruieren.

ecli.ordinalzahl <- paste0(gsub("Bv([A-Z])",
                                "\\1",
                                txt.bverfg$registerzeichen),
                           txt.bverfg$spruchkoerper_typ,
                           txt.bverfg$datum,
                           txt.bverfg$kollision,
                           ".",
                           txt.bverfg$spruchkoerper_az,
                           txt.bverfg$registerzeichen,
                           formatC(txt.bverfg$eingangsnummer,
                                   width = 4,
                                   flag = "0"),
                           formatC(txt.bverfg$eingangsjahr_az,
                                   width = 2,
                                   flag = "0"))

ecli.ordinalzahl <- gsub("NA",
                         "",
                         ecli.ordinalzahl)

ecli.ordinalzahl <- gsub("-",
                         "",
                         ecli.ordinalzahl)

ecli.ordinalzahl <- gsub("vzb",
                         "vb",
                         ecli.ordinalzahl)

ecli.ordinalzahl <- gsub("pup",
                         "up",
                         ecli.ordinalzahl)


ecli.ordinalzahl <- tolower(ecli.ordinalzahl)


txt.bverfg$ecli <- paste0("ECLI:DE:BVerfG:",
                          txt.bverfg$entscheidungsjahr,
                          ":",
                          ecli.ordinalzahl)



#'## Variable "doi_concept" hinzufügen

txt.bverfg$doi_concept <- rep(doi.concept,
                              txt.bverfg[,.N])


#'## Variable "doi_version" hinzufügen

txt.bverfg$doi_version <- rep(doi.version,
                              txt.bverfg[,.N])


#'## Variable "version" hinzufügen

txt.bverfg$version <- as.character(rep(datestamp,
                                       txt.bverfg[,.N]))







#'# Frequenztabellen erstellen

#+
#'## Funktion anzeigen

#+ results = "asis"
print(f.fast.freqtable)


#'## Ignorierte Variablen
print(varremove)



#'## Liste zu prüfender Variablen

varlist <- names(txt.bverfg)
varlist <- grep(paste(varremove,
                      collapse="|"),
                varlist,
                invert = TRUE,
                value = TRUE)
print(varlist)



#'## Frequenztabellen erstellen

prefix <- paste0(datasetname,
                 "_01_Frequenztabelle_var-")


#+ results = "asis"
f.fast.freqtable(txt.bverfg,
                 varlist = varlist,
                 sumrow = TRUE,
                 output.list = FALSE,
                 output.kable = TRUE,
                 output.csv = TRUE,
                 outputdir = outputdir,
                 prefix = prefix)





#'# Frequenztabellen visualisieren

#'## Präfix erstellen

prefix <- paste0("ANALYSE/",
                 datasetname,
                 "_01_Frequenztabelle_var-")


#'## Tabellen einlesen

table.spruch.typ <- fread(paste0(prefix,
                                 "spruchkoerper_typ.csv"))

table.spruch.az <- fread(paste0(prefix,
                                "spruchkoerper_az.csv"))

table.regz <- fread(paste0(prefix,
                           "registerzeichen.csv"))

table.jahr.eingangISO <- fread(paste0(prefix,
                                      "eingangsjahr_iso.csv"))

table.jahr.entscheid <- fread(paste0(prefix,
                                     "entscheidungsjahr.csv"))

table.output.praesi <- fread(paste0(prefix,
                                    "praesi.csv"))

table.output.vpraesi <- fread(paste0(prefix,
                                     "v_praesi.csv"))





#'\newpage
#'## Spruchkörper Typ

freqtable <- table.spruch.typ[-.N]


#+ CE-BVerfG_02_Barplot_Spruchkoerper_Typ, fig.height = 5, fig.width = 8
ggplot(data = freqtable) +
    geom_bar(aes(x = reorder(spruchkoerper_typ, N),
                 y = N),
             stat = "identity",
             fill ="#ca2129",
             color = "black",
             width = 0.4) +
    theme_bw() +
    labs(
        title = paste(datasetname,
                      "| Version",
                      datestamp,
                      "| Entscheidungen je Spruchkörper-Typ"),
        caption = paste("DOI:",
                        doi.version),
        x = "Typ des Spruchkörpers",
        y = "Entscheidungen"
    )+
    theme(
        text = element_text(size=14),
        plot.title = element_text(size=14, face="bold"),
        legend.position="none",
        plot.margin = margin(10, 20, 10, 10)
    )




#'\newpage
#'## Spruchkörper AZ

freqtable <- table.spruch.az[-.N]


#+ CE-BVerfG_03_Barplot_Spruchkoerper_AZ, fig.height = 5, fig.width = 8
ggplot(data = freqtable) +
    geom_bar(aes(x = spruchkoerper_az,
                 y = N),
             stat = "identity",
             fill ="#ca2129",
             color = "black",
             width = 0.4) +
    theme_bw() +
    labs(
        title = paste(datasetname,
                      "| Version",
                      datestamp,
                      "| Entscheidungen je Senat (Aktenzeichen)"),
        caption = paste("DOI:",
                        doi.version),
        x = "Senat",
        y = "Entscheidungen"
    )+
    theme(
        text = element_text(size=14),
        plot.title = element_text(size=14, face="bold"),
        legend.position="none",
        plot.margin = margin(10, 20, 10, 10)
    )



#'\newpage
#'## Registerzeichen

freqtable <- table.regz[-.N]

#+ CE-BVerfG_04_Barplot_Registerzeichen, fig.height = 10, fig.width = 8
ggplot(data = freqtable) +
    geom_bar(aes(x = reorder(registerzeichen, N),
                 y = N),
             stat="identity",
             fill ="#ca2129",
             color = "black") +
    coord_flip()+
    theme_bw() +
    labs(
        title = paste(datasetname,
                      "| Version",
                      datestamp,
                      "| Entscheidungen je Registerzeichen"),
        caption = paste("DOI:",
                        doi.version),
        x = "Registerzeichen",
        y = "Entscheidungen"
    )+
    theme(
        text = element_text(size=14),
        plot.title = element_text(size=14, face="bold"),
        legend.position="none",
        plot.margin = margin(10, 20, 10, 10)
    )



#'\newpage
#'## PräsidentIn

freqtable <- table.output.praesi[-.N]

#+ CE-BVerfG_05_Barplot_PraesidentIn, fig.height = 5.5, fig.width = 8
ggplot(data = freqtable) +
    geom_bar(aes(x = reorder(praesi, N),
                 y = N),
             stat="identity",
             fill ="#ca2129",
             color = "black") +
    coord_flip()+
    theme_bw() +
    labs(
        title = paste(datasetname,
                      "| Version",
                      datestamp,
                      "| Entscheidungen je PräsidentIn"),
        caption = paste("DOI:",
                        doi.version),
        x = "PräsidentIn",
        y = "Entscheidungen"
    )+
    theme(
        axis.title.y=element_blank(),
        text = element_text(size=14),
        plot.title = element_text(size=14, face="bold"),
        legend.position="none",
        plot.margin = margin(10, 20, 10, 10)
    )



#'\newpage
#'## Vize-PräsidentIn

freqtable <- table.output.vpraesi[-.N]

#+ CE-BVerfG_06_Barplot_VizePraesidentIn, fig.height = 5.5, fig.width = 8
ggplot(data = freqtable) +
    geom_bar(aes(x = reorder(v_praesi, N),
                 y = N),
             stat="identity",
             fill ="#ca2129",
             color = "black") +
    coord_flip()+
    theme_bw() +
    labs(
        title = paste(datasetname,
                      "| Version",
                      datestamp,
                      "| Entscheidungen je Vize-PräsidentIn"),
        caption = paste("DOI:",
                        doi.version),
        x = "Vize-PräsidentIn",
        y = "Entscheidungen"
    )+
    theme(
        axis.title.y=element_blank(),
        text = element_text(size=14),
        plot.title = element_text(size=14, face="bold"),
        legend.position="none",
        plot.margin = margin(10, 20, 10, 10)
    )



#'\newpage
#'## Entscheidungsjahr

freqtable <- table.jahr.entscheid[-.N][,lapply(.SD, as.numeric)]

#+ CE-BVerfG_07_Barplot_Entscheidungsjahr, fig.height = 7, fig.width = 11
ggplot(data = freqtable) +
    geom_bar(aes(x = entscheidungsjahr,
                 y = N),
             stat="identity",
             fill ="#ca2129") +
    theme_bw() +
    labs(
        title = paste(datasetname,
                      "| Version",
                      datestamp,
                      "| Entscheidungen je Entscheidungsjahr"),
        caption = paste("DOI:",
                        doi.version),
        x = "Entscheidungsjahr",
        y = "Entscheidungen"
    )+
    theme(
        text = element_text(size=16),
        plot.title = element_text(size=16, face="bold"),
        legend.position="none",
        plot.margin = margin(10, 20, 10, 10)
    )



#'\newpage
#'## Eingangsjahr (ISO)

freqtable <- table.jahr.eingangISO[-.N][,lapply(.SD, as.numeric)]



#+ CE-BVerfG_08_Barplot_EingangsjahrISO, fig.height = 7, fig.width = 11
ggplot(data = freqtable) +
    geom_bar(aes(x = eingangsjahr_iso,
                 y = N),
             stat="identity",
             fill ="#ca2129") +
    theme_bw() +
    labs(
        title = paste(datasetname,
                      "| Version",
                      datestamp,
                      "| Entscheidungen je Eingangsjahr (ISO)"),
        caption = paste("DOI:",
                        doi.version),
        x = "Eingangsjahr (ISO)",
        y = "Entscheidungen"
    )+
    theme(
        text = element_text(size=16),
        plot.title = element_text(size=16, face="bold"),
        legend.position="none",
        plot.margin = margin(10, 20, 10, 10)
    )





#'# Korpus-Analytik
#'

#+
#'## Berechnung linguistischer Kennwerte
#' An dieser Stelle werden für jedes Dokument die Anzahl Tokens, Typen und Sätze berechnet und mit den jeweiligen Metadaten verknüpft. Das Ergebnis ist grundsätzlich identisch mit dem eigentlichen Datensatz, nur ohne den Text der Entscheidungen.

corpus <- corpus(txt.bverfg)

result <- foreach(i = seq_len(ndoc(corpus)),
                  .errorhandling = 'pass') %dopar% {
                      temp <- summary(corpus[i])
                      return(temp)
               }

summary.corpus <- rbindlist(result)


#'## Variablen-Namen anpassen

setnames(summary.corpus,
         old = c("Text",
                 "Tokens",
                 "Types",
                 "Sentences"),
         new = c("doc_id",
                 "tokens",
                 "typen",
                 "saetze"))


#'## Kennwerte dem Korpus hinzufügen

txt.bverfg$tokens <- summary.corpus$tokens
txt.bverfg$typen <- summary.corpus$typen
txt.bverfg$saetze <- summary.corpus$saetze


#'## Anzahl Variablen im Korpus
length(txt.bverfg)


#'## Alle Variablen-Namen im Korpus
names(txt.bverfg)



#'## Linguistische Kennwerte
#' **Hinweis:** Typen sind definiert als einzigartige Tokens und werden für jedes Dokument gesondert berechnet. Daher ergibt es an dieser Stelle auch keinen Sinn die Typen zu summieren, denn bezogen auf den Korpus wäre der Kennwert ein anderer.  Der Wert wird daher manuell auf "NA" gesetzt.

#+
#'### Zusammenfassungen berechnen

dt.summary.ling <- summary.corpus[, lapply(.SD,
                                           function(x)unclass(summary(x))),
                                  .SDcols = c("tokens",
                                              "saetze",
                                              "typen")]


dt.sums.ling <- summary.corpus[,
                               lapply(.SD, sum),
                               .SDcols = c("tokens",
                                           "saetze",
                                           "typen")]

dt.sums.ling$typen <- NA




dt.stats.ling <- rbind(dt.sums.ling,
                       dt.summary.ling)

dt.stats.ling <- transpose(dt.stats.ling,
                           keep.names = "names")


setnames(dt.stats.ling, c("Variable",
                          "Sum",
                          "Min",
                          "Quart1",
                          "Median",
                          "Mean",
                          "Quart3",
                          "Max"))



#'### Zusammenfassungen anzeigen

kable(dt.stats.ling,
      format.args = list(big.mark = ","),
      format = "latex",
      booktabs=TRUE,
      longtable=TRUE)




#'### Zusammenfassungen speichern

fwrite(dt.stats.ling,
       paste0(outputdir,
              datasetname,
              "_00_KorpusStatistik_ZusammenfassungLinguistisch.csv"),
       na = "NA")





#'## Quantitative Variablen


#'### Entscheidungsdatum

summary(as.IDate(summary.corpus$datum))



#'### Zusammenfassungen berechnen

dt.summary.docvars <- summary.corpus[,
                                     lapply(.SD, function(x)unclass(summary(na.omit(x)))),
                                     .SDcols = c("entscheidungsjahr",
                                                 "eingangsjahr_iso",
                                                 "band",
                                                 "eingangsnummer")]


dt.unique.docvars <- summary.corpus[,
                                    lapply(.SD, function(x)length(unique(na.omit(x)))),
                                    .SDcols = c("entscheidungsjahr",
                                                "eingangsjahr_iso",
                                                "band",
                                                "eingangsnummer")]


dt.stats.docvars <- rbind(dt.unique.docvars,
                          dt.summary.docvars)

dt.stats.docvars <- transpose(dt.stats.docvars,
                              keep.names = "names")


setnames(dt.stats.docvars, c("Variable",
                             "Unique",
                             "Min",
                             "Quart1",
                             "Median",
                             "Mean",
                             "Quart3",
                             "Max"))



#'### Zusammenfassungen anzeigen

kable(dt.stats.docvars,
      format = "latex",
      booktabs=TRUE,
      longtable=TRUE)



#'### Zusammenfassungen speichern

fwrite(dt.stats.docvars,
       paste0(outputdir,
              datasetname,
              "_00_KorpusStatistik_ZusammenfassungDocvarsQuantitativ.csv"),
       na = "NA")




#'\newpage
#'## Density

#+
#'### Density Tokens

#+ CE-BVerfG_09_Density_Tokens, fig.height = 6, fig.width = 9
ggplot(data = summary.corpus)+
    geom_density(aes(x = tokens),
                 fill = "#ca2129")+
    scale_x_log10(breaks = trans_breaks("log10", function(x) 10^x),
                  labels = trans_format("log10", math_format(10^.x)))+
    annotation_logticks(sides = "b")+
    coord_cartesian(xlim = c(1, 10^6))+
    theme_bw()+
    labs(
        title = paste(datasetname,
                      "| Version",
                      datestamp,
                      "| Verteilung der Tokens je Dokument"),
        caption = paste("DOI:",
                        doi.version),
        x = "Tokens",
        y = "Dichte"
    )+
    theme(
        text = element_text(size=14),
        plot.title = element_text(size=14, face="bold"),
        legend.position="none",
        plot.margin = margin(10, 20, 10, 10)
    )



#'\newpage
#'### Density Typen

#+ CE-BVerfG_10_Density_Typen, fig.height = 6, fig.width = 9
ggplot(data = summary.corpus)+
    geom_density(aes(x = typen),
                 fill = "#ca2129")+
    scale_x_log10(breaks = trans_breaks("log10", function(x) 10^x),
                  labels = trans_format("log10", math_format(10^.x)))+
    annotation_logticks(sides = "b")+
    coord_cartesian(xlim = c(1, 10^6))+
    theme_bw()+
    labs(
        title = paste(datasetname,
                      "| Version",
                      datestamp,
                      "| Verteilung der Typen je Dokument"),
        caption = paste("DOI:",
                        doi.version),
        x = "Typen",
        y = "Dichte"
    )+
    theme(
        text = element_text(size=14),
        plot.title = element_text(size=14, face="bold"),
        legend.position="none",
        plot.margin = margin(10, 20, 10, 10)
    )

#'\newpage
#'### Density Sätze

#+ CE-BVerfG_11_Density_Saetze, fig.height = 6, fig.width = 9
ggplot(data = summary.corpus)+
    geom_density(aes(x = saetze),
                 fill = "#ca2129")+
    scale_x_log10(breaks = trans_breaks("log10", function(x) 10^x),
                  labels = trans_format("log10", math_format(10^.x)))+
    annotation_logticks(sides = "b")+
    coord_cartesian(xlim = c(1, 10^6))+
    theme_bw()+
    labs(
        title = paste(datasetname,
                      "| Version",
                      datestamp,
                      "| Verteilung der Sätze je Dokument"),
        caption = paste("DOI:",
                        doi.version),
        x = "Sätze",
        y = "Dichte"
    )+
    theme(
        text = element_text(size=14),
        plot.title = element_text(size=14, face="bold"),
        legend.position="none",
        plot.margin = margin(10, 20, 10, 10)
    )



#'# Beispiel-Werte für alle Metadaten anzeigen

print(summary.corpus)



#'# CSV-Dateien erstellen

#'## CSV mit vollem Datensatz speichern

csvname.full <- paste(datasetname,
                      datestamp,
                      "DE_CSV_Datensatz.csv",
                      sep = "_")

fwrite(txt.bverfg,
       csvname.full,
       na = "NA")




#'## CSV mit Metadaten speichern
#' Diese Datei ist grundsätzlich identisch mit dem eigentlichen Datensatz, nur ohne den Text der Entscheidungen.

csvname.meta <- paste(datasetname,
                      datestamp,
                      "DE_CSV_Metadaten.csv",
                      sep = "_")

fwrite(summary.corpus,
       csvname.meta,
       na = "NA")







#'# Dateigrößen analysieren

#+
#'## Gesamtgröße

#'### Korpus-Objekt in RAM (MB)

print(object.size(corpus),
      standard = "SI",
      humanReadable = TRUE,
      units = "MB")


#'### CSV Korpus (MB)
file.size(csvname.full) / 10 ^ 6


#'### CSV Metadaten (MB)
file.size(csvname.meta) / 10 ^ 6


#'### PDF-Dateien (MB)

files.pdf <- list.files(pattern = "\\.pdf$",
                        ignore.case = TRUE)

pdf.MB <- file.size(files.pdf) / 10^6
sum(pdf.MB)


#'### TXT-Dateien (MB)
files.txt <- list.files(pattern = "\\.txt$",
                        ignore.case = TRUE)

txt.MB <- file.size(files.txt) / 10^6
sum(txt.MB)





#'\newpage
#'## Verteilung der Dateigrößen (PDF)

dt.plot <- data.table(pdf.MB)

#+ CE-BVerfG_12_Density_Dateigroessen, fig.height = 7, fig.width = 9
ggplot(data = dt.plot,
       aes(x = pdf.MB)) +
    geom_density(fill ="#ca2129") +
    scale_x_log10(breaks = trans_breaks("log10", function(x) 10^x),
                  labels = trans_format("log10", math_format(10^.x)))+
    annotation_logticks(sides = "b")+
    theme_bw() +
    labs(
        title = paste(datasetname,
                      "| Version",
                      datestamp,
                      "| Verteilung der Dateigrößen (PDF)"),
        caption = paste("DOI:",
                        doi.version),
        x = "Dateigröße in MB",
        y = "Dichte"
    )+
    theme(
        text = element_text(size=14),
        plot.title = element_text(size=14, face="bold"),
        legend.position="none",
        panel.spacing = unit(0.1, "lines"),
        plot.margin = margin(10, 20, 10, 10)
    )




#'# Erstellen der ZIP-Archive

#'## Verpacken der CSV-Dateien

#+ results = 'hide'
csvname.full.zip <- gsub(".csv",
                    ".zip",
                    csvname.full)

zip(csvname.full.zip,
    csvname.full)

unlink(csvname.full)


#+ results = 'hide'
csvname.meta.zip <- gsub(".csv",
                    ".zip",
                    csvname.meta)

zip(csvname.meta.zip,
    csvname.meta)




#'## Verpacken der PDF-Dateien

#+ results = 'hide'
zip(paste(datasetname,
          datestamp,
          "DE_PDF_Datensatz.zip",
          sep = "_"),
    files.pdf)

unlink(files.pdf)


#'## Verpacken der TXT-Dateien

#+ results = 'hide'
files.txt <- list.files(pattern="\\.txt",
                        ignore.case = TRUE)

zip(paste(datasetname,
          datestamp,
          "DE_TXT_Datensatz.zip",
          sep = "_"),
    files.txt)

unlink(files.txt)


#'## Verpacken der Analyse-Dateien

zip(paste0(datasetname,
           "_",
           datestamp,
           "_DE_",
           basename(outputdir),
           ".zip"),
    basename(outputdir))




#'## Verpacken der Source-Dateien

files.source <- c(list.files(pattern = "Source"),
                  "buttons")


files.source <- grep("spin",
                     files.source,
                     value = TRUE,
                     ignore.case = TRUE,
                     invert = TRUE)

zip(paste(datasetname,
           datestamp,
           "Source_Files.zip",
           sep = "_"),
    files.source)





#'# Kryptographische Hashes
#' Dieses Modul berechnet für jedes ZIP-Archiv zwei Arten von Hashes: SHA2-256 und SHA3-512. Mit diesen kann die Authentizität der Dateien geprüft werden und es wird dokumentiert, dass sie aus diesem Source Code hervorgegangen sind. Die SHA-2 und SHA-3 Algorithmen gelten derzeit als sicher und ein SHA3-Hash mit 512 bit Länge ist nach derzeitigem Wissen auch gegenüber quantenkryptoanalytischen Verfahren hinreichend resistent.

#+
#'## Liste der ZIP-Archive erstellen
files.zip <- list.files(pattern= "\\.zip$",
                        ignore.case = TRUE)


#'## Funktion anzeigen
#+ results = "asis"
print(f.dopar.multihashes)

#'## Hashes berechnen
multihashes <- f.dopar.multihashes(files.zip)


#'## In Data Table umwandeln
setDT(multihashes)



#'## Index hinzufügen
multihashes$index <- seq_len(multihashes[,.N])


#'## In Datei schreiben
fwrite(multihashes,
       paste(datasetname,
             datestamp,
             "KryptographischeHashes.csv",
             sep = "_"),
       na = "NA")


#'## Leerzeichen hinzufügen um Zeilenumbruch zu ermöglichen
multihashes$sha3.512 <- paste(substr(multihashes$sha3.512, 1, 64),
                              substr(multihashes$sha3.512, 65, 128))


#'\newpage
#'## In Bericht anzeigen

kable(multihashes[,.(index,filename)],
      format = "latex",
      align = c("p{1cm}",
                "p{13cm}"),
      booktabs=TRUE,
      longtable=TRUE)


kable(multihashes[,.(index,sha2.256)],
      format = "latex",
      align = c("c",
                "p{13cm}"),
      booktabs=TRUE,
      longtable=TRUE)


#'\newpage

kable(multihashes[,.(index,sha3.512)],
      format = "latex",
      align = c("c",
                "p{13cm}"),
      booktabs=TRUE,
      longtable=TRUE)







#'# Abschluss

#'## Cluster stoppen
stopCluster(cl)


#'## Datum und Uhrzeit (Ende)
end.script <- Sys.time()
print(end.script)

#'## Laufzeit des gesamten Skriptes
print(end.script - begin.script)


#'## Warnungen
warnings()



#'# Parameter für strenge Replikationen

system2("openssl", "version", stdout = TRUE)

sessionInfo()


#'# Literaturverzeichnis

